#'Create mask array from a shapefile (.shp) and a dataset, and save to netCDF 
#'file (optionally)
#' 
#'A shapefile is the file type saving the information of polygonal regions. 
#'First the function reads the .shp file and transfers it to an array. Then, 
#'we subset the output for the requested region names or IDs. This information 
#'is different from each shapefile format. The accepted shapefiles databases 
#'names are: 'NUTS', 'LAU' and 'GADM'. After this step, we load the reference 
#'dataset to access the longitude and latitude in order to compare the shapefile 
#'and the coordinates. After loading the coordinates we intersect each subset 
#'of the shapefile with them in order to select only the desired regions. Then, 
#'the function creates a mask array. In the last step, the mask is either 
#'returned or saved into NetCDF format in the desired directory.
#'
#'Modules GDAL, PROJ and GEOS are required.
#' 
#'@param shp.file A character string indicating the shp file path.
#'@param ref.grid A character string indicating the path to the reference 
#'  data. Either (1) a netCDF file or (2) a list of lon and lat to provide the 
#'  reference grid points. It is NULL by default.
#'@param shp.system It is set to 'NUTS' by default.
#'@param reg.ids A character string indicating The unique ID in shapefile. 
#'  It is NULL by default.
#'@param reg.names A named list of character string vectors indicating the 
#'  country and the region name. The name of the list stands for the country 
#'  name code and the vector character strings indicate the region name for 
#'  each country. It is NULL by default.
#'@param reg.level An integer number from 1 to 3 indicating the 'NUTS' dataset 
#'  level. For other datasets this parameter is not used. One mask can only have 
#'  a unique level. It is set to 3 by default.
#'@param lat_dim A character string indicating the latitudinal dimension. If it 
#'  is NULL, the latitudinal name will be searched using an internal function 
#'  with the following possible names: 'lat', 'latitude', 'y', 'j' and 
#'  'nav_lat'. It is NULL by default.
#'@param lon_dim A character string indicating the longitudinal dimension. If it 
#'  is NULL, the longitudinal name will be searched using an internal function 
#'  with the following possible names: 'lon', 'longitude', 'x', 'i' and 
#'  'nav_lon'. It is NULL by default.
#'@param savefile A logical value indicating wether to save the mask array into 
#'  a NetCDF format file (TRUE) or to return an array (FALSE). It is FALSE by 
#'  default.
#' 
#'@examples 
#'# Exmple (1): NUTS
#'#-------shp.file----------
#'shp.file <- paste0('/esarchive/shapefiles/NUTS3/NUTS_RG_60M_2021_4326.shp/', 
#'                   'NUTS_RG_60M_2021_4326.shp')
#'shp.file <- paste0('/esarchive/scratch/cdelgado/focus_outputs/Shapefiles/',
#'                   'tza_admbnda_adm1/tza_admbnda_adm1_20181019.shp')
#'#-------ref.grid----------
#'ref.grid <- paste0('/esarchive/exp/ecmwf/system5c3s/monthly_mean/',
#'                   'tas_f6h/tas_20170201.nc')
#'ref.grid <- paste0('/esarchive/exp/ecmwf/s2s-monthly_ensfor/weekly_mean/',
#'                   'tas_f6h/tas_20191212.nc')
#'ref.grid <- paste0('/esarchive/recon/ecmwf/era5land/monthly_mean/',
#'                   'tas_f1h/tas_201006.nc')
#'ref.grid <- list(lon = seq(10, 40, 0.5), lat = seq(40, 85, 0.5))
#'#-------reg.ids-----------
#'NUTS.id <- paste0("FI1D", c(1:3, 5, 7:9))
#'# ADM.id <- 
#'#-------reg.names-----------
#'# NOTE: NUTS.name may be repetitive; use level to filter
#'NUTS.name <- list(FI = c('Lappi', 'Kainuu'))   
#'NUTS.name <- list(FI = c('Lappi', 'Kainuu'), SI = c('Pomurska', 'Podravska'))
#'#---------------------------
#'mask1 <- shp_mask(shp.file, ref.grid, reg.ids = NUTS.id)
#'mask2 <- shp_mask(shp.file, ref.grid, reg.names = NUTS.name)

#'# Exmple (2): GADM
#'#-------shp.file----------
#'shp.file <- "/esarchive/shapefiles/gadm_country_mask/gadm_country_ISO3166.shp"
#'#-------ref.grid----------
#'ref.grid <- paste0('/esarchive/exp/ecmwf/s2s-monthly_ensfor/weekly_mean/',
#'                   'tas_f6h/tas_20191212.nc')
#-------reg.ids-----------
#'GADM.id <- c("ESP", "ITA")
#-------reg.names-----------
#'GADM.name <- c("Spain", "Italy")
#---------------------------
#'mask1 <- shp_mask(shp.file = shp.file, ref.grid = ref.grid, 
#'                  reg.ids = GADM.id, shp.system = "GADM")
#'mask2 <- shp_mask(shp.file = shp.file, ref.grid = ref.grid, 
#'                  reg.names = GADM.name, shp.system = "GADM")
#'@import easyNCDF
#'@import sf
#'@export
shp_mask <- function(shp.file, ref.grid = NULL, 
                     shp.system = "NUTS", reg.ids = NULL, reg.names = NULL,
                     reg.level = 3, lat_dim = NULL, lon_dim = NULL, savefile = FALSE) {
#NOTE: One region is one number; need to have the option to combine them?
#TODO: Suppress warnings?
#TODO: Substitute packages
####################################################
  # Step 1: Load the shapefile
  shp <- sf::st_read(shp.file)  # class sf
 
  if (all(is.null(reg.ids), is.null(reg.names))) {
    stop("Either provide parameter 'reg.ids' or 'reg.names'.")
  
  } else if (!is.null(reg.ids)) {
    ## Method 1: Directly use IDs
    if (shp.system == "NUTS") {
      shp <- subset(shp, NUTS_ID %in% reg.ids)
    } else if (shp.system == "ADM") {
      shp <- subset(shp, ADM1_PCODE %in% reg.ids)
    } else if (shp.system == "GADM") {
      shp <- subset(shp, ISO %in% reg.ids)
    } else {
      stop("shp.system ", shp.system, " is not defined yet.")
    }
    if (!is.null(reg.names)) {
      warning("Only use 'reg.ids' to get the shape region. 'reg.names' is not used.")
    }
  
  } else if (!is.null(reg.names)) {
    ## Method 2: Use country code & region name
    for (cntr_i in 1:length(reg.names)) {
      if (shp.system == "NUTS") {
        tmp <- subset(shp, CNTR_CODE == names(reg.names)[cntr_i])
        tmp <- subset(tmp, NUTS_NAME %in% reg.names[[cntr_i]])
      } else if (shp.system == "ADM") {
        tmp <- subset(shp, ADM0_EN == names(reg.names)[cntr_i])
        tmp <- subset(tmp, ADM1_EN %in% reg.names[[cntr_i]])
      } else if (shp.system == "GADM") {
        tmp <- subset(shp, Name %in% reg.names)
      }
      if (cntr_i == 1) {
        shp_tmp <- tmp
      } else {
        shp_tmp <- rbind(shp_tmp, tmp)
      }
    }
    if (shp.system == "NUTS") {
      shp <- subset(shp_tmp, LEVL_CODE == reg.level)
    } else if (shp.system == "ADM" | shp.system == "GADM") {
      shp <- shp_tmp
    }
  }
  # plot(shp)
  
####################################################
  # Step 2: Use the reference file to get lat and lon
 
  if (all(tools::file_ext(ref.grid) == 'nc')) {
    if (!file.exists(ref.grid)) {
      stop("ref.grid file does not exist.")
    } else {
    
      ## Method 1: ref.grid is a netCDF file
      if (is.null(lat_dim) | is.null(lon_dim)) {
        var_names <- easyNCDF::NcReadVarNames(ref.grid)
        lat_dim <- var_names[which(var_names %in% s2dv:::.KnownLatNames())]
        lon_dim <- var_names[which(var_names %in% s2dv:::.KnownLonNames())]
      }
      latlon <- NcToArray(ref.grid, vars_to_read = c(lat_dim, lon_dim))
      lat <- NcToArray(ref.grid, vars_to_read = lat_dim)[1, ]
      lon <- NcToArray(ref.grid, vars_to_read = lon_dim)[1, ]
    }
  } else if (is.list(ref.grid)) {
    ## Method 2: ref.grid is a list of lon and lat
    if (length(ref.grid) != 2) {
      stop("If 'ref.grid' is a list, it must have two items for longitude and latitude.")
    }
    if (is.null(lat_dim) | is.null(lon_dim)) {
      #NOTE: the names come from s2dv:::.KnownLonNames and .KnownLatNames
      lon_known_names <- c(s2dv:::.KnownLonNames(), 'lons')
      lat_known_names <- c(s2dv:::.KnownLatNames(), 'lats')
      lon_dim <- lon_known_names[which(lon_known_names %in% names(ref.grid))]
      lat_dim <- lat_known_names[which(lat_known_names %in% names(ref.grid))]

      if (identical(lon_dim, character(0)) | identical(lat_dim, character(0))) {
        stop("longitude and latitude names are not recognized in 'ref.grid'. Please specify 'lon_dim' and 'lat_dim'.")
      }
    }
    lat <- ref.grid[[lat_dim]]
    lon <- ref.grid[[lon_dim]]

  } else {
    stop("Parameter 'ref.grid' must be either a netCDF file or a list of lon and lat.")
  }

  ## Create data frame & sp class for ref grid
  ref.df <- data.frame(data = 0,
                       lon = rep(lon, times = length(lat)),
                       lat = rep(lat, each = length(lon)))
  coord <- as.matrix(data.frame(x = ref.df$lon, y = ref.df$lat))
  
  xy.sfg <- st_multipoint(coord)
  xy.sfc <- st_sfc(xy.sfg)
  st_crs(xy.sfc) <- sf::st_crs(shp)

####################################################
  # Step 3: Create mask 
  ## Create mask array with 0; 1, 2, etc. will be filled in for each shp region
  mask <- array(0, dim = c(length(lon), length(lat)))
  names(dim(mask)) <- c(lon_dim, lat_dim)
  
  ## Loop through each shp region
  for (shp_i in 1:nrow(shp)) {
    #NOTE: Don't know it's a problem in st_intersection or st_coordinates, tmp_coords may 
    #      not be identical as lon/lat. E.g., (29, 65) may be (29 - -3.552714e-15, 65).
    shp_pol <- sf::st_intersection(xy.sfc, shp[shp_i, ])
    tmp_coords <- st_coordinates(shp_pol)[, 1:2]
    if (nrow(tmp_coords) != 0) {
      for (ii in 1:nrow(tmp_coords)) {
        pt_x <- which(lon == tmp_coords[ii, 1])
        pt_y <- which.min(abs(lat - tmp_coords[ii, 2]))
        mask[which.min(abs(lon - tmp_coords[ii, 1])),
             which.min(abs(lat - tmp_coords[ii, 2]))] <- shp_i
        #NOTE: This line can't catch correct point when the NOTE above happens.
        #mask[which(lon == tmp_coords[ii, 1]), which(lat == tmp_coords[ii, 2])] <- shp_i
      }
    } else {
      if (shp.system == "NUTS") {
        tmp <- shp$NUTS_ID[shp_i]
      } else if (shp.system == "ADM") {
        tmp <- shp$ADM1_PCODE[shp_i]
      } else if (shp.system == "GADM") {
        tmp <- shp$ISO[shp_i]
      }
      warning("shp ID '", tmp, "' cannot be identified in this reference grid.")
    }
  }
  
####################################################
  # Step 4: Add attributes
  attr(mask, lon_dim) <- lon
  attr(mask, lat_dim) <- lat
  if (shp.system == "NUTS") {
    attr(mask, "index") <- as.list(shp$NUTS_ID)
  } else if (shp.system == "ADM") {
    attr(mask, "index") <- as.list(shp$ADM1_PCODE)
  } else if (shp.system == "GADM") {
    attr(mask, "index") <- as.list(shp$ISO)
  }
  names(attr(mask, "index")) <- 1:nrow(shp)

  ## Return all the info from shp
  attr(mask, "shapefile") <- attributes(shp)
  
####################################################
  # Step 5: Save the file or return the array 
  if (!savefile) {
    return(mask)
  } else {
    warning("This functionality is not developed yet.")
    #TODO
    # ArrayToNc()
  }
}

